/* Threedler
 *
 * Lunar Lander AI written by Brad Smith, 3/07/12
 *
 * 1. Continually builds a pascal style floodfill from the goal to the lander,
 *    with a priority queue evaluating new nodes closer to the lander first.
 *    When the lander is reached, the path is immediately rebuilt, so
 *    each pass will take into account information gained since the last.
 * 2. Each frame evaluates several potential paths to take,
 *    takes the one that crashes in the least amount of time, then
 *    reaches the best pathing point, then reaches that point in the least
 *    amount of time.
 * 3. When the goal is reached, takes a very simple approach to landing,
 *    borrowed from twodler. (not very good)
 *
 * Ran out of time on this one, there's one or more bugs
 * This might do worse than the twodler because of the bugs.
 *
 * http://rainwarrior.ca
 */

#include "ai_interface.h"
#include <cmath>
#include <queue>
#include <vector>

// Constants
// =========

// debug constants

// enables debug lines
#define DEBUG_DRAW 0
// enables keyboard override with cursor keys (down toggles override)
#define KEYBOARD_TEST 0
// enables LOG macro (printf style debug output)
#define LOGGING 0

// simulation/math constants

const float PI = 3.1415926536f;
const float ROOT2 = ::sqrtf(2.0f);
const float RT2O2 = ROOT2 / 2.0f;

const float GRAV      = float(GRAVITY);
const float TIMESTEP  = 0.01f;
const float ROT_SPEED = PI; // radians per second
const float THRUST    = 5.0f * GRAV;

// control constants

const float LOOK_AHEAD     = 4.0f;  // total time evaluated for each branch
const float CRASH_IGNORE   = 3.0f;  // ignore crashes if they are this far ahead
const int   THRUST_TESTS   = 4;     // resolution of lookahead thrust branching
const float THRUST_TIME    = 1.0f;  // time to apply thrust before evaluating
const float MAX_VEL        = 10.0f; // reaching this velocity treated as crash
const float TARGET_OFFSET  = 5.0f;  // aim this distance above landing point
const float LAND_DIST      = 3.0f;  // start landing this close to target
const float CLOUD_RES      = 1.0f;  // resolution of point cloud
const int   JOBS_PER_FRAME = 50;    // number of pathing jobs to evaluate per frame
const unsigned int MAX_JOBS = 16000; // restart if max jobs reached
const int   FRAME_SKIP     = 1;     // frames per AI change

// debug visualization constants

const unsigned int COLOR_LOCATION = 0xFFFFFF00;
const unsigned int COLOR_VELOCITY = 0xFF00FF00;
const unsigned int COLOR_SENSOR   = 0xFF0000FF;
const unsigned int COLOR_FLAT     = 0xFF008855;
const float CLOUD_DRAW_DIST = 25.0f; // distance to draw cloud around lander

// structures

typedef unsigned __int32 uint32;
typedef unsigned __int8  uint8;

struct LanderState
{
    float x, y;
    float vx, vy;
    float rot;
};

const uint32 CLOUD_UNKNOWN = 0x80000000; // means point is totally unknown
const uint32 CLOUD_BLOCK   = 0x40000000; // means known point is either blocked or unblocked
const uint32 CLOUD_PASS    = 0x3FFF0000; // pass counter (lower passes are better)
const uint32 CLOUD_VALUE   = 0x0000FFFF; // last evaluated value mask
const uint32 CLOUD_CRASH   = 0xFFFFFFFF; // worst possible value

struct SeekEval
{
    float time_to_crash; // time before crash found
    uint32 best_cloud;   // best cloud value passed through
    float best_time;     // time to best cloud value
};

struct PathJob
{
    int x, y;
    uint32 priority;
    bool operator<(const PathJob &p) const { return priority > p.priority; }
    bool operator>(const PathJob &p) const { return priority < p.priority; }
};

// global variables

LanderState current;
float goalx;
float goaly;
bool landing;
unsigned int frame_skip = 0;

int cloud_w = 800;
int cloud_h = 800;
float cloud_min_x = -300.0f;
float cloud_min_y = -300.0f;
float cloud_max_x = float(cloud_w) * CLOUD_RES;
float cloud_max_y = float(cloud_h) * CLOUD_RES;
uint32* cloud = NULL;
uint32 cloud_pass = CLOUD_PASS >> 16;
unsigned int job_count = MAX_JOBS;

std::priority_queue<PathJob> path_queue;

uint32 arc_color = 0;

// Debug stuff
// ===========

#if KEYBOARD_TEST || LOGGING
    #define WIN32_LEAN_AND_MEAN
    #include <windows.h>
#endif

bool log_on = false;
#if LOGGING
    #define LOG(...) \
    { \
        if (log_on) \
        { \
            char temp_[256]; \
            ::sprintf_s(temp_, sizeof(temp_), __VA_ARGS__ ); \
            ::OutputDebugStringA(temp_); \
        } \
    }
#else
    #define LOG(...) {}
#endif

// Geometry: it's the Law.
// =======================

float length(float x, float y)
{
    return ::sqrtf((x*x)+(y*y));
}

void normalize(float x, float y, float* ox, float* oy)
{
    float l = length(x,y);
    if (l <= 0.0f)
    {
        *ox = 0.0f;
        *oy = 0.0f;
    }
    *ox = x / l;
    *oy = y / l;
}

void direction(float angle, float* x, float* y)
{
    *x = -::sinf(angle);
    *y = ::cosf(angle);
}

inline int iabs(int x)
{
    return x >= 0 ? x : -x;
}

// Lander state simulator
// ======================

// single step simulation
LanderState simulate_step(LanderState l, bool thrust, bool rotl, bool rotr)
{
    if (rotl) l.rot += ROT_SPEED * TIMESTEP;
    if (rotr) l.rot -= ROT_SPEED * TIMESTEP;

    l.vy -= GRAV * TIMESTEP;

    if (thrust)
    {
        float dx, dy;
        direction(l.rot, &dx, &dy);
        l.vx += dx * THRUST * TIMESTEP;
        l.vy += dy * THRUST * TIMESTEP;
    }

    l.x += l.vx * TIMESTEP;
    l.y += l.vy * TIMESTEP;

    return l;
}

// multi step simulation
LanderState simulate_time(LanderState l, bool thrust, bool rotl, bool rotr, float t)
{
    float arot = l.rot;
    if (rotl) l.rot += ROT_SPEED * t;
    if (rotr) l.rot -= ROT_SPEED * t;
    arot = (arot + l.rot) * 0.5f; // approximate with average rotation

    float ax, ay;
    if (thrust)
    {
        direction(arot, &ax, &ay);
        ax *= THRUST;
        ay *= THRUST;
    }
    else
    {
        ax = 0.0f;
        ay = 0.0f;
    }
    ay -= GRAV;

    // premultiply acceleration by time / 2
    ax *= 0.5f * t;
    ay *= 0.5f * t;

    l.x += (l.vx + ax) * t; // pos += vel * t + 1/2 accel * t^2
    l.y += (l.vy + ay) * t;
    l.vx += ax;
    l.vy += ay;

    return l;
}

// Collision cloud
// ===============

bool cloud_bounds(int x, int y)
{
    return (x >= 0 && y >= 0 && x < cloud_w && y < cloud_h);
}

uint32& cloud_array(int x, int y)
{
    if (cloud_bounds(x,y))
    {
        return cloud[(y*cloud_w)+x];
    }

    static uint32 cloud_outside = CLOUD_CRASH;
    return cloud_outside;
}

// TODO resizable cloud

void world_to_cloud(float fx, float fy, int* ox, int* oy)
{
    fx -= cloud_min_x;
    fy -= cloud_min_y;
    fx /= (cloud_max_x - cloud_min_x);
    fy /= (cloud_max_y - cloud_min_y);
    fx *= float(cloud_w);
    fy *= float(cloud_h);
    *ox = (int)fx;
    *oy = (int)fy;
}

void cloud_to_world(int ix, int iy, float* ox, float* oy)
{
    *ox = (float(ix) * (cloud_max_x - cloud_min_x) / float(cloud_w)) + cloud_min_x;
    *oy = (float(iy) * (cloud_max_y - cloud_min_y) / float(cloud_h)) + cloud_min_y;
}

// add point to cloud
void cloud_add(float x, float y)
{
    // mark blocked
    int ax, ay;
    world_to_cloud(x, y, &ax, &ay);
    if (cloud_bounds(ax, ay))
    {
        cloud_array(ax,ay) = CLOUD_CRASH;
    }

    // mark known all points from lander to here
    int lx, ly;
    world_to_cloud(current.x, current.y, &lx, &ly);
    int dx = (ax == lx) ? 0 : ((ax > lx) ? 1 : -1);
    int dy = (ay == ly) ? 0 : ((ay > ly) ? 1 : -1);
    int ix = lx;
    int iy = ly;
    while (ix != ax || iy != ay)
    {
        uint32 val = cloud_array(ix,iy);
        if (val == CLOUD_CRASH)
        {
            break; // don't erase known crashes, just give up
        }
        val = val ^ (val & CLOUD_UNKNOWN);
        cloud_array(ix,iy) = val;

        if (ax != ix) ix += dx;
        if (ay != iy) iy += dy;
    }
}

uint32 cloud_read(float x, float y)
{
    int ix, iy;
    world_to_cloud(x, y, &ix, &iy);
    if (!cloud_bounds(ix, iy))
    {
        return CLOUD_CRASH; // treat off-cloud as a crash
    }

    uint32 val = cloud_array(ix,iy);

    // crash points are the worst possible value
    if (val == CLOUD_CRASH)
        return CLOUD_CRASH;

    // if already evaluated, just return the value
    if (0 == (val & CLOUD_UNKNOWN) && ((val & CLOUD_PASS) != CLOUD_PASS))
        return val;

    // point is unevaluated, generate an estimate
    float fx, fy;
    cloud_to_world(ix, iy, &fx, &fy);
    uint32 estimate = uint32(length(fx-goalx,fy-goaly));
    val &= CLOUD_UNKNOWN | CLOUD_PASS;
    val |= (estimate & CLOUD_VALUE);
    return val;
}

void cloud_draw(float flx, float fly, float frx, float fry)
{
    if(!DEBUG_DRAW) return;

    int lx,ly,rx,ry;
    world_to_cloud(flx,fly,&lx,&ly);
    world_to_cloud(frx,fry,&rx,&ry);
    for (int iy = ly; iy <= ry; ++iy)
    for (int ix = lx; ix <= rx; ++ix)
    {
        if(cloud_bounds(ix,iy))
        {
            float fx, fy;
            cloud_to_world(ix,iy,&fx,&fy);
            uint32 c = cloud_read(fx,fy);

            float blx, bly, brx, bry;
            cloud_to_world(ix  , iy  , &blx, &bly);
            cloud_to_world(ix+1, iy+1, &brx, &bry);

            uint32 v = c & CLOUD_VALUE;
            //uint32 v1 = v > 255 ? 255 : v;
            //uint32 v2 = v > 127 ? 255 : (v * 2);
            //uint32 color = 0xFF000000 | (v1 << 16) | (v2 << 8) | v2;
            uint32 color = 0xFFFFFFFF;
            if (c == CLOUD_CRASH) color = 0xFFFFFFFF;
            else
            {
                if (c & CLOUD_UNKNOWN) continue;
                uint32 p = (c & CLOUD_PASS) >> 16;
                if      (p ==     cloud_pass) color &= 0xFF00FF00;
                else if (p == (cloud_pass+1)) color &= 0xFFFFFF00;
                else                          color &= 0xFF0000FF;
            }

            Lunar_AI::Draw_Line(blx,bly,brx,bry,false,color);
            Lunar_AI::Draw_Line(brx,bly,blx,bry,false,color);
        }
    }
}

// Cloud pathing data generation
// =============================

void cloud_init()
{
    cloud = new uint32[cloud_w * cloud_h];
    for (int i=0; i < (cloud_w * cloud_h); ++i)
    {
        cloud[i] = CLOUD_UNKNOWN | CLOUD_PASS;
    }
    cloud_pass = CLOUD_PASS >> 16;
    while (!path_queue.empty())
    {
        path_queue.pop();
    }
    job_count = MAX_JOBS;
}

void cloud_cleanup()
{
    delete [] cloud;
    cloud = NULL;
}

// cloud_tick builds a path to the player
void cloud_tick()
{
    cloud_draw(
        current.x - CLOUD_DRAW_DIST,
        current.y - CLOUD_DRAW_DIST,
        current.x + CLOUD_DRAW_DIST,
        current.y + CLOUD_DRAW_DIST);

    LOG("cloud_tick()\n");

    // neighbour nodes for easy loop
    const int dx[4] = { -1, 1, 0, 0 };
    const int dy[4] = { 0, 0, -1, 1 };

    int gx, gy;
    world_to_cloud(goalx, goaly, &gx, &gy);

    int lx, ly;
    world_to_cloud(current.x, current.y, &lx, &ly);
    const float lfx = float(lx);
    const float lfy = float(ly);

    // begin a new pass if the queue is empty
    // or if we've already reached the lander
    if(path_queue.empty() ||
       ((cloud_read(current.x,current.y) & CLOUD_PASS) == (cloud_pass << 16))
       || (job_count < 1)
       )
    {
        if (cloud_pass > 0)
        {
            --cloud_pass;
        }
        else
        {
            // run out of passes, just reset
            for (int iy=0; iy < cloud_h; ++iy)
            for (int ix=0; ix < cloud_w; ++ix)
            {
                cloud_array(ix,iy) |= CLOUD_PASS;
            }
            cloud_pass = CLOUD_PASS - 1;
        }
        // resart the pass
        while (!path_queue.empty())
        {
            path_queue.pop();
        }
        LOG("New pass: %4x\n", cloud_pass);
        PathJob job = { gx, gy, 0 };
        path_queue.push(job);
        job_count = MAX_JOBS;
    }

    uint32 pass = cloud_pass << 16;

    unsigned int jobs = JOBS_PER_FRAME;
    while (!path_queue.empty() && jobs > 0 && job_count > 0)
    {
        --jobs;
        --job_count;
        PathJob job = path_queue.top();
        path_queue.pop();

        //LOG("Job: %4d, %4d = %8d\n", job.x, job.y, job.priority);

        const int x = job.x;
        const int y = job.y;
        if (!cloud_bounds(x,y)) continue; // out of bounds

        uint32 c = cloud_array(x,y);
        if (c == CLOUD_CRASH) continue; // don't path here
        if ((c & CLOUD_PASS) == pass) continue; // already evaluated

        // base transition cost to reach point,
        // favour known empty points
        uint32 cost = (0 == (c & CLOUD_UNKNOWN)) ? 1 : 2;

        // add best neighbour cost to transition cost
        uint32 best_neighbour = CLOUD_VALUE;
        for (unsigned int i=0; i < 4; ++i)
        {
            uint32 cn = cloud_array(x+dx[i],y+dy[i]);
            if ((cn & CLOUD_PASS) == pass) // point has been evaluated
            {
                uint32 vn = cn & CLOUD_VALUE;
                if (vn < best_neighbour)
                {
                    best_neighbour = vn;
                }
            }
        }
        if (best_neighbour == CLOUD_VALUE)
        {
            // orhpan node? discard if not the starting node
            if (x == gx && y == gy)
            {
                //LOG("Starting node.\n");
                best_neighbour = 0;
            }
            else
            {
                //LOG("Orphaned node, skipping.\n");
                continue;
            }
        }
        cost += best_neighbour;

        // store value
        //LOG("Cost: %4d, %4d = %8d\n", x, y, cost);
        if (cost > CLOUD_VALUE) cost = CLOUD_VALUE;
        c = (c & CLOUD_UNKNOWN);
        cloud_array(x,y) = c | pass | (cost & CLOUD_VALUE);

        if (job.x == lx && job.y == ly)
        {
            // lunar lander has been reached, so
            // clear the queue, start a new pass next frame
            while (!path_queue.empty())
            {
                path_queue.pop();
            }
        }
        else
        {
            // push adjacent jobs
            for (unsigned int i=0; i < 4; ++i)
            {
                int xn = x + dx[i];
                int yn = y + dy[i];

                float fx, fy;
                cloud_to_world(xn, yn, &fx, &fy);
                uint32 estimate = uint32(length(fx-current.x, fy-current.y));
                uint32 priority = cost + estimate;
                PathJob job = { xn, yn, priority };
                path_queue.push(job);
            }
        }
    }
}

// Seeking
// =======

// returns true if a is better than b
bool seek_better(SeekEval a, SeekEval b)
{
    if (a.time_to_crash > b.time_to_crash && b.time_to_crash < CRASH_IGNORE) return true;
    if (a.time_to_crash < b.time_to_crash && a.time_to_crash < CRASH_IGNORE) return false;
    if (a.best_cloud    < b.best_cloud   ) return true;
    if (a.best_cloud    > b.best_cloud   ) return false;
    if (a.best_time     < b.best_time    ) return true;
    return false;
}

void seek_eval_arc(SeekEval& e, LanderState& l, bool thrust, float t, float st)
{
    // calculate a 3 point now -> peak -> end segment
    float peak = l.vy / GRAV;
    if (peak < 0.0f) peak = 0.0f;
    if (peak > t) peak = t;
    LanderState lp = simulate_time(l,thrust,false,false,peak);
    LanderState le = simulate_time(l,thrust,false,false,t);
    if (DEBUG_DRAW)
    {
        //Lunar_AI::Draw_Line(l.x,l.y,lp.x,lp.y,false,COLOR_FLAT | arc_color);
        //Lunar_AI::Draw_Line(lp.x,lp.y,le.x,le.y,false,COLOR_FLAT | arc_color);
    }
    // use 3-point path to estimate segments needed to make sure each timestep
    // is small enough to be shorter than CLOUD_RES
    float flat_len = length(lp.x-l.x, lp.y-l.y) + length(le.x-lp.x, le.y-lp.y);
    const float CURVE_ADJUST = 1.5f;
    int segments = 1 + int(flat_len * CURVE_ADJUST / CLOUD_RES);

    LanderState la = l;
    for (int segment = 0; segment <= segments; ++segment)
    {
        float seg_t = t * float(segment)/float(segments);
        LanderState lb = simulate_time(l,thrust,false,false,seg_t);
        if (DEBUG_DRAW)
            Lunar_AI::Draw_Line(la.x, la.y, lb.x, lb.y, false, arc_color);
        la = lb;

        // total time to this segment
        float accum_t = seg_t + st;

        // current cloud point tells us how "good" this point is
        uint32 c = cloud_read(la.x, la.y);

        LOG("seek_eval_arc %d: %8.4f %08x = %8.4f %08x %8.4f\n",
            thrust, accum_t, c,
            e.time_to_crash, e.best_cloud, e.best_time
            );

        // exceeding maximum velocity is treated as crash
        if (length(la.vx, la.vy) >= MAX_VEL)
            c = CLOUD_CRASH;

        // don't venture into the unknown
        if (c & CLOUD_UNKNOWN)
            c = CLOUD_CRASH;

        // if we crash, stop evaluating
        if (c == CLOUD_CRASH)
        {
            e.time_to_crash = accum_t;
            break;
        }

        // if this point is better than other points reached so far, remember it
        if (c < e.best_cloud)
        {
            e.best_cloud = c;
            e.best_time = accum_t;
        }
    }

    // move lander to last reached point
    l = la;
}


// evaluate the future of a set of options
SeekEval seek_eval(LanderState l, bool thrust, bool rotl, bool rotr, float t, float et)
{
    LOG("seek_eval %d|%d|%d: %8.4f %8.4f\n", thrust, rotl, rotr, t, et);

    uint32 c = cloud_read(l.x, l.y);
    SeekEval base_eval;
    base_eval.time_to_crash = t;
    base_eval.best_cloud = cloud_read(l.x, l.y);
    base_eval.best_time = 0.0f;

    float turn = (rotl ? 1.0f : 0.0f) + (rotr ? -1.0f : 0.0f);
    float rend = l.rot + (turn * et * ROT_SPEED);
    float rmid = (rend + l.rot) * 0.5f;

    // take early step applying options
    l.rot = rmid; // approximate with halfway rotation
    seek_eval_arc(base_eval, l, thrust, et, 0.0f);
    l.rot = rend; // finish rotation
    if (thrust && et < THRUST_TIME)
    {
        seek_eval_arc(base_eval, l, true, THRUST_TIME-et, et);
        et = THRUST_TIME;
    }

    // if we've already crashed, we're done
    if (base_eval.time_to_crash < t)
    {
        return base_eval;
    }
    float time_left = t - et;

    // test proceeding from here without rotating any more,
    // but thrusting for various amounts of the time
    SeekEval best_eval = base_eval;
    for (unsigned int i=0; i < THRUST_TESTS; ++i)
    {
        float time_thrust = time_left * float(i) / float(THRUST_TESTS);
        float time_float = time_left - time_thrust;

        SeekEval eval = base_eval;
        LanderState lt = l;
        LOG("seek_eval: thrust %d/%d\n", i, THRUST_TESTS);
        seek_eval_arc(eval, lt, true, time_thrust, et);
        seek_eval_arc(eval, lt, false, time_float, et);

        if (seek_better(eval, best_eval) || i==0)
        {
            best_eval = eval;
        }
    }

    // return best possible outcome from evaluated paths
    return best_eval;
}

// try to navigate to goal_x/goal_y
void seek(bool &out_thrust, bool &out_left, bool &out_right)
{
    LOG("seek()\n");

    // fill "best" with worst possible outcome (immediate crash)
    bool best_thrust = false;
    bool best_left   = false;
    bool best_right  = false;
    SeekEval best;
    int besti, bestj;
    best.time_to_crash = 0.0f;
    best.best_cloud = CLOUD_CRASH;
    best.best_time = LOOK_AHEAD;

    static float TEST_SPREAD = 0.6f * (PI / 2.0f);
    static const int TEST_ROT_NUM = 5;
    static const float TEST_ROT[TEST_ROT_NUM] = {
        -(1.0f * TEST_SPREAD),
        -(0.5f * TEST_SPREAD),
          0.0f,
         (0.5f * TEST_SPREAD),
                 TEST_SPREAD
        };

    for (int j=0; j < 2; ++j)
    for (int i=0; i < TEST_ROT_NUM; ++i)
    {
        if (DEBUG_DRAW)
        {
            arc_color = 0xFF00FF00 | (((((i*2)+j) * 255) / (2*TEST_ROT_NUM)) << 16);
        }

        float aim = TEST_ROT[i];
        float rot_time = ::fabsf(current.rot - aim) / ROT_SPEED;

        bool test_thrust = (j & 1) ? true : false;
        bool test_left  = (aim > current.rot) ? true : false;
        bool test_right = (aim < current.rot) ? true : false;

        SeekEval eval = seek_eval(
            current,
            test_thrust, test_left, test_right,
            LOOK_AHEAD,
            rot_time);

        LOG("eval %d|%d: %8.4f %08x %8.4f (%d %d %d)\n",
            i,j,
            eval.time_to_crash, eval.best_cloud, eval.best_time,
            test_thrust, test_left, test_right
            );

        if (seek_better(eval, best))
        {
            best = eval;
            besti = i;
            bestj = j;
            best_thrust = test_thrust;
            best_left   = test_left;
            best_right  = test_right;
        }
    }

    if (DEBUG_DRAW)
    {
        arc_color = 0xFFFF0000;

        float aim = TEST_ROT[besti];
        float rot_time = ::fabsf(current.rot - aim) / ROT_SPEED;

        bool test_thrust = (bestj & 1) ? true : false;
        bool test_left  = (aim > current.rot) ? true : false;
        bool test_right = (aim < current.rot) ? true : false;

        SeekEval eval = seek_eval(
            current,
            test_thrust, test_left, test_right,
            LOOK_AHEAD,
            rot_time);
    }

    LOG("Best: %d|%d (%d %d %d)\n", besti, bestj, best_thrust, best_thrust, best_right);

    // HACK disabling thrust if we're gonna crash
    if (best.time_to_crash <= 0.0f)
    {
        out_thrust = false;
    }

    // apply best option
    out_thrust = best_thrust;
    out_left   = best_left;
    out_right  = best_right;
}

// Landing
// =======

// quickly taken wholesale from lunar_twodler

const float MAX_LAND_FALL = 1.0f;
const float MAX_LAND_VX = 1.0f;

enum Turn {
    TURN_LEFT,
    TURN_RIGHT,
    TURN_UP
    };

bool turning = false;
Turn turn_goal = TURN_UP;

void land_set_turn(Turn t)
{
    turning = true;
    turn_goal = t;
}

// just turn
bool land_turn(bool& out_thrust, bool& out_left, bool& out_right)
{
    const float EPS = 1.1f  * ROT_SPEED;

    switch (turn_goal)
    {
        case TURN_LEFT:
            if (current.rot < (PI *  0.25f))
                out_left = true;
            else
                turning = false;
            break;
        case TURN_RIGHT:
            if (current.rot > (PI * -0.25f))
                out_right = true;
            else
                turning = false;
            break;
        case TURN_UP:
            if (current.rot > EPS)
                out_right = true;
            else if (current.rot < -EPS)
                out_left = true;
            else
                turning = false;
            break;
    }
    return turning;
}

void land(bool &out_thrust, bool &out_left, bool &out_right)
{
    out_left = false;
    out_right = false;

    if (turning)
    {
        land_turn(out_thrust, out_left, out_right);
    }

    Turn t = TURN_UP;
    if (current.x < goalx || current.vx < -MAX_LAND_VX)
        t = TURN_RIGHT;
    if (current.x > goalx || current.vx > MAX_LAND_VX)
        t = TURN_LEFT;

    if (t != TURN_UP && t != turn_goal)
        land_set_turn(t);

    if (current.vy < -MAX_LAND_FALL)
        out_thrust = true;
}

// Init/cleanup
// ============

void init()
{
    cloud_init();
    landing = false;
    frame_skip = 0;

    turning = false;
    turn_goal = TURN_UP;
}

void cleanup()
{
    cloud_cleanup();
}

// KEYBOARD_TEST
// =============

void keyboard_test(bool &out_thrust, bool &out_left, bool &out_right)
{
    #if (KEYBOARD_TEST)
    {
        static bool keyboard_test_override = false;

        // if down was pressed this frame, toggle override
        static bool toggle_key = false;
        bool toggle_key_on = (::GetKeyState(VK_DOWN) < 0);
        if (toggle_key_on && !toggle_key)
        {
            keyboard_test_override = !keyboard_test_override;
        }
        toggle_key = toggle_key_on;

        if (keyboard_test_override)
        {
            out_thrust = ::GetKeyState(VK_UP) < 0;
            out_left   = ::GetKeyState(VK_LEFT) < 0;
            out_right  = ::GetKeyState(VK_RIGHT) < 0;
        }

        if (LOGGING)
        {
            static bool log_key = false;
            bool log_key_on = ::GetKeyState(VK_F12) < 0;
            if (log_key_on & !log_key)
            {
                log_on = !log_on;
            }
            log_key = log_key_on;
        }
    }
    #endif
}

// DLL framework
// =============

namespace Lunar_AI
{
    const char * Get_AI_Name()
    {
        return "threedler";
    }

    void Init()
    {
        LOG("Init()\n");
        init();
    }

    void Cleanup()
    {
        LOG("Cleanup()\n");
        cleanup();
    }

    void Tick( float dt, float x, float y, float vx, float vy, float rot,
                float gx, float gy,
                float sense[ NUM_SENSORS ], float fuel,
                bool &out_thrust, bool &out_left, bool &out_right)
    {
        // store current state globally

        current.x = x;
        current.y = y;
        current.vx = vx;
        current.vy = vy;
        current.rot = rot;

        goalx = gx;
        goaly = gy;

        // visualize known information

        if (DEBUG_DRAW)
        {
            Draw_Text(x - 2.0f, y - 2.0f, 0.5f, false, COLOR_LOCATION, "(%6.2f, %6.2f)", x, y);
            Draw_Text(x - 2.0f, y - 2.5f, 0.5f, false, COLOR_VELOCITY, "(%6.2f, %6.2f) %6.2f", vx, vy, length(vx, vy));
            Draw_Line(x, y, x+vx, y+vy, false, COLOR_VELOCITY);
        }

        // build point cloud from sensor data

        float sx[NUM_SENSORS];
        float sy[NUM_SENSORS];
        static const float SDIR[NUM_SENSORS][2] = {
            { -RT2O2,  RT2O2 },
            {   0.0f,   1.0f },
            {  RT2O2,  RT2O2 },
            {  -1.0f,   0.0f },
            {   1.0f,   0.0f },
            { -RT2O2, -RT2O2 },
            {   0.0f,  -1.0f },
            {  RT2O2, -RT2O2 } };

        for (int i=0; i < NUM_SENSORS; ++i)
        {
            sx[i] = x + (SDIR[i][0] * sense[i]);
            sy[i] = y + (SDIR[i][1] * sense[i]);
            if (sense[i] < SENSOR_MAX_RANGE)
            {
                cloud_add(sx[i], sy[i]);
            }

            if (DEBUG_DRAW)
            {
                Draw_Line(x, y, sx[i], sy[i], false, COLOR_SENSOR);
            }
        }

        if (landing) // landing
        {
            land(out_thrust, out_left, out_right);
        }
        else // seeking
        {
            goaly += TARGET_OFFSET; // seek a target above the actual landing point

            cloud_tick();

            static bool last_thrust = false;
            static bool last_left   = false;
            static bool last_right  = false;
            if (frame_skip == 0)
            {
                seek(out_thrust, out_left, out_right);
                if (length(x - goalx, y - goaly) <= LAND_DIST)
                {
                    // begin landing when you get close
                    landing = true;
                }
                frame_skip = FRAME_SKIP;
                last_thrust = out_thrust;
                last_left   = out_left;
                last_right  = out_right;
            }
            else
            {
                out_thrust = last_thrust;
                out_left   = last_left;
                out_right  = last_right;
            }
            --frame_skip;
        }

        // keyboard override
        keyboard_test(out_thrust, out_left, out_right);
    }

}// namespace Lunar_AI
